#!/usr/bin/env python
# -*- coding: utf-8 -*-

from typing import List
from pathlib import Path

import numpy as np
import h5py
from tqdm import tqdm

from .common_tools import get_zaptime_files, save_create, measure_speed
from .read_edf import read_edf_gz
from .gisaxs_parameters import GisaxsParameters

__all__ = ['h5_manager']


class H5PathNotProvidedError(ValueError):
    """
    Please, set h5 _path first.
    """
    pass


class H5PathIsNotAFileError(ValueError):
    """
    Please, check the h5 file _path you provided.
    """
    pass


class ParentDirNotExistsError(ValueError):
    """
    The path is invalid. Please, provide a path to an existing folder.
    """


class H5Manager(object):
    @property
    def path(self):
        return self._path

    @property
    def xpcs_path(self):
        return self._xpcs_path

    def __init__(self, path: str = None, xpcs_path: str = None):
        self._path = path
        self._xpcs_path = xpcs_path

    def add_files(self, zaptime_number: List[int] or int):
        if not self._path:
            raise H5PathNotProvidedError()
        with h5py.File(self._path, 'a') as h5file:
            add_raw_data_to_h5(h5file, zaptime_number,
                               save_axes=True)

    def delete_files(self, zaptime_number: List[int] or int):
        if not self._path:
            raise H5PathNotProvidedError()
        with h5py.File(self._path, 'a') as h5file:
            delete_data_from_h5(h5file, zaptime_number)

    def set_path(self, path: str or Path):
        if isinstance(path, str):
            path = Path(path)
        if not path.is_file():
            raise H5PathIsNotAFileError()
        self._path = path

    def set_xpcs_path(self, xpcs_path: str or Path):
        if isinstance(xpcs_path, str):
            xpcs_path = Path(xpcs_path)
        if not xpcs_path.is_file():
            if xpcs_path.parent.is_dir():
                with h5py.File(xpcs_path, 'w'):
                    pass
            else:
                raise ParentDirNotExistsError()
        self._xpcs_path = xpcs_path


@measure_speed('Adding raw data to h5 file')
def add_raw_data_to_h5(h5file: h5py.File, zaptime_number: List[int] or int, *,
                       save_axes: bool = True, rewrite: bool = False) -> None:
    """
    Adds raw data from zaptime folder on groupshare with chosen number to the target h5 file.
    If save_axes is True, it adds axes as datasets and creates reference in attrs dict.

    :param h5file: target h5 file
    :param zaptime_number: target zaptime folder number
    :param save_axes: (optional) if True, saves axes as datasets to h5 file.
    :param rewrite: (optional) if True, allows to rewrite existing groups in h5file.
    :return: None
    """
    if isinstance(zaptime_number, list):
        for num in zaptime_number:
            add_raw_data_to_h5(h5file, num, save_axes=save_axes, rewrite=rewrite)
    else:
        print('Folder number ', zaptime_number)
        gp = GisaxsParameters.init_from_folder_number(zaptime_number)
        if save_axes:
            x_axis, y_axis = gp.get_angle_vectors(units='deg', flip=False)
        else:
            x_axis = y_axis = None
        zaptime_files = get_zaptime_files(zaptime_number)
        if zaptime_files:
            raw_data_group = save_create(h5file, 'Raw data')
            current_group_name = '_'.join(['zaptime', str(zaptime_number).zfill(2)])
            if current_group_name in list(raw_data_group.keys()) and not rewrite:
                print('Current folder already exists')
                return
            current_group = save_create(raw_data_group, current_group_name)
            if x_axis is not None and y_axis is not None:
                x_axis_dset = current_group.create_dataset('x_axis', data=x_axis)
                y_axis_dset = current_group.create_dataset('y_axis', data=y_axis)
            else:
                x_axis_dset = y_axis_dset = None
            for i, gz_file in enumerate(tqdm(zaptime_files)):
                try:
                    data, header_dict = read_edf_gz(gz_file)
                    data = np.flip(data, axis=1)
                    dset = current_group.create_dataset(str(i).zfill(len(str(len(zaptime_files)))), data=data)
                    dset.attrs.update(header_dict)
                    if x_axis_dset is not None and y_axis_dset is not None:
                        dset.attrs.update({'x_axis': x_axis_dset.ref, 'y_axis': y_axis_dset.ref})
                except Exception as err:
                    print(err)


@measure_speed('Delete data from h5')
def delete_data_from_h5(h5file: h5py.File, zaptime_number: List[int] or int):
    if isinstance(zaptime_number, list):
        for num in zaptime_number:
            delete_data_from_h5(h5file, num)
    else:
        if 'Raw data' not in h5file.keys():
            print('Raw data not found!')
            return
        else:
            group = h5file['Raw data']
            current_group_name = '_'.join(['zaptime', str(zaptime_number).zfill(2)])
            if current_group_name not in group.keys():
                print(f'Zaptime folder {zaptime_number} not found.')
                return
            else:
                del group[current_group_name]


h5_manager = H5Manager()

if __name__ == '__main__':
    folder_number = 76
    path = '/media/vladimir/data/GISAXS_DATA/test_raw_data.h5'
    with h5py.File(path, 'a') as h5file:
        add_raw_data_to_h5(h5file, folder_number)
        # delete_data_from_h5(h5file, list(range(76)))



# -*- coding: utf-8 -*-

import pytest
from gisaxs_xpcs import h5_manager, MONETA_DIR

__all__ = ['moneta_connnection', 'h5_raw_data_path', 'h5_xpcs']


@pytest.fixture(scope='session', autouse=True)
def moneta_connnection():
    MONETA_DIR._set_default_dir()


@pytest.fixture(scope='session', autouse=True)
def h5_raw_data_path():
    h5_manager.set_path('/media/vladimir/data1/GISAXS_DATA/test_raw_data.h5')


@pytest.fixture(scope='session', autouse=True)
def h5_xpcs():
    h5_manager.set_xpcs_path('/media/vladimir/data1/GISAXS_DATA/xpcs_data.h5')



# -*- coding: utf-8 -*-
import pytest
from gisaxs_xpcs import h5_manager, MONETA_DIR


def test_moneta_connection():
    assert MONETA_DIR.path.is_dir()


def test_h5_path():
    assert h5_manager.path.is_file()


def test_xpcs_path():
    assert h5_manager.xpcs_path.parent.is_dir()



#!/usr/bin/env python
# -*- coding: utf-8 -*-


import os
import logging
from typing import Tuple

import numpy as np
import h5py

from ..gisaxs_parameters import GisaxsParameters
from ..manage_h5_data import h5_manager

from .calculate_xpcs import calculate_1d_corr, calculate_2d_corr, XPCSModes
from .extract_intensities import get_intensities
from .region_grid import RegionType

logger = logging.getLogger(__name__)


class XPCS(object):
    """
        Class provides the basic functionality to calculate
        two-time correlation functions for
        GISAXS images stored to .h5 file.
    """

    @staticmethod
    def _default_h5_path(folder_num: int):
        return '/'.join(['Raw data', 'zaptime_%s' % str(folder_num).zfill(2)])

    def __init__(self, folder_num: int = None,
                 cut_region: RegionType = None,
                 gp: GisaxsParameters = None,
                 h5_filepath: str = None,
                 h5_path: str = None,
                 mask: RegionType = None):
        """

        :param folder_num: number of the zaptime folder to analyse
        :param cut_region: tuple(x_min, x_max, y_min, y_max) in [deg]
        :param gp: gisaxs parameters object (necessary to extract angles)
        :param h5_filepath: _path to the data source
        :param h5_path: _path to zaptime folder in h5 file
        """
        if not folder_num and not h5_filepath:
            raise ValueError('Either folder_num or h5_filename should be provided')
        self.folder_num = folder_num
        self._index_x = self._index_y = None
        self._mask_x = self._mask_y = None
        self._bool_mask = None
        if not gp and folder_num:
            self.gp = GisaxsParameters.init_from_folder_number(folder_num)
        else:
            self.gp = gp or GisaxsParameters()
        self._cut_region = cut_region
        self._mask = mask
        self.h5_filepath = h5_filepath or h5_manager.path
        self.h5_path = h5_path or self._default_h5_path(self.folder_num)
        if not self._setup_h5():
            raise NotImplementedError('Reading from remote folder is not'
                                      ' implemented yet.')

        self.xx, self.yy = self._init_mask_coords()

        self._update_mask()

    @property
    def mask(self):
        return self._bool_mask

    @property
    def region(self):
        return self._cut_region

    @property
    def status(self) -> bool:
        return self.mask is not None

    def set_region(self, cut_region: RegionType):
        self._cut_region = cut_region
        self._update_mask()

    def set_mask(self, mask: RegionType):
        self._mask = mask
        self._update_mask()

    def get_two_time_corr_function(self, mode: str = 'std') -> Tuple[np.ndarray, np.ndarray]:
        if not self.status:
            raise ValueError('Cut region should be provided.')
        if mode == 'mean':
            mode = XPCSModes.mean
        elif mode == 'std':
            mode = XPCSModes.std
        else:
            raise ValueError(f'Mode should be "std" or "mean", {mode} provided.')
        intensity, time_array = self.get_intensities()
        return calculate_2d_corr(intensity, mode), time_array

    def get_one_time_corr_function(self) -> Tuple[np.ndarray, np.ndarray]:
        if not self.status:
            raise ValueError('Cut region should be provided.')
        intensity, time_array = self.get_intensities()
        return calculate_1d_corr(intensity), time_array

    def get_intensities(self) -> Tuple[np.ndarray, np.ndarray]:
        return get_intensities(self.h5_filepath, self.mask, self.h5_path)

    def _setup_h5(self) -> bool:
        """
        Checks .h5 file. If it exists and contains the provided _path to the dataset,
        returns True, otherwise False.
        """
        if not os.path.isfile(self.h5_filepath):
            logger.error('File %s doesn\'t exist.' % self.h5_filepath)
            return False
        try:
            with h5py.File(self.h5_filepath, 'r') as f:
                if f.get(self.h5_path, None) is None:
                    logger.error('File does not contain _path %s' % self.h5_path)
                    return False
                # data = f[self.h5_path][list(f[self.h5_path].keys())[0]][()]
                # print(data.shape)
        except Exception as err:
            logger.exception(err)
            return False
        return True

    def _init_mask_coords(self):
        x_angles, y_angles = self.gp.get_angle_vectors(units='deg',
                                                       flip=False)
        x, y = np.sort(x_angles), np.flip(np.sort(y_angles), axis=0)
        return np.meshgrid(x, y)

    def _update_mask(self) -> None:
        if not self._cut_region:
            mask = np.ones_like(self.xx, dtype=bool)
        else:
            mask = self._get_mask(self._cut_region)
        if self._mask is not None:
            mask = mask & self.beam_stop_mask
        self._bool_mask = mask

    @property
    def beam_stop_mask(self):
        if self._mask is not None:
            return np.logical_not(self._get_mask(self._mask))
        else:
            return np.ones_like(self.xx, dtype=bool)

    def _get_mask(self, region: RegionType):
        return (self.xx > region[2]) & (self.xx < region[3]) & (self.yy > region[0]) * (self.yy < region[1])



# -*- coding: utf-8 -*-

from typing import (Tuple, Iterable, Union,
                    List, Callable, Dict)

import numpy as np
from scipy.ndimage import zoom
from h5py import File
from tqdm import tqdm
import matplotlib.pyplot as plt
import matplotlib.patches as patches

from ..manage_h5_data import h5_manager
from ..extract_data import ExtractData

from .xpcs import XPCS
from .region_grid import (RegionGrid, RGridParams,
                          RListParams, RegionType)

GeneratorType = Iterable[Tuple[np.ndarray, RegionType]]


class XPCSMap(object):
    @property
    def parameters(self) -> Union[RListParams, RGridParams]:
        return self.region_grid.parameters

    @staticmethod
    def _default_xpcs_map_h5_group_name(folder_num: int) -> str:
        return 'zaptime_%d' % folder_num

    def __init__(self, folder_num: int,
                 region_grid: RegionGrid,
                 mask: RegionType = None,
                 name: str = 'xpcs_map'):
        self.folder_num = folder_num
        self.region_grid = region_grid
        self.name = name
        self.xpcs = XPCS(folder_num, mask=mask)

    def get_time_axis(self):
        _, t_array = self.xpcs.get_intensities()
        return t_array

    def set_name(self, name: str) -> None:
        self.name = name

    def generate_2d_corr(self, mode: str = 'std') -> GeneratorType:
        for region in self.region_grid:
            self.xpcs.set_region(region)
            corr, _ = self.xpcs.get_two_time_corr_function(mode)
            yield corr, region

    def generate_1d_corr(self) -> GeneratorType:
        for region in self.region_grid:
            self.xpcs.set_region(region)
            corr, _ = self.xpcs.get_one_time_corr_function()
            yield corr, region

    def get_one_2d_corr_from_file(self, region: RegionType):
        res = _get_regions_from_h5(self.folder_num, self.name, [region])
        try:
            return res[region]
        except KeyError:
            raise ValueError(f'Region {region} not found in saved xpcs.')

    def save_2d_corr(self, mode: str = 'std') -> None:
        _save_to_h5(self.name, self.folder_num,
                    self.generate_2d_corr(mode), self.parameters)

    def generate_from_file(self, **kwargs) -> GeneratorType:
        return _read_from_h5(self.folder_num, self.name, **kwargs)

    def plot_grid(self, file_num: int = 0, **kwargs) -> None:
        extract_data = ExtractData()
        extract_data.get_images(
            self.folder_num, file_num, apply_log=True, plot=True,
            show=False, gp=self.xpcs.gp, x_range=(-10, 10), y_range=(-10, 10)
        )

        ax = plt.gca()
        for cut in self.region_grid:
            ax.add_patch(_get_rect(cut, **kwargs))

        plt.show()

    def plot_from_file(self, figname: str = None,
                       figsize: Tuple[int, int] = None,
                       reduce_factor: float = None,
                       reduce_mode: str = 'nearest') -> None:
        generator = self.generate_from_file()
        _plot_r(generator, self.parameters,
                reduce_factor, reduce_mode, figsize)
        if figname:
            plt.savefig(figname)
        else:
            plt.show()


def _plot_r(generator: GeneratorType, parameters: RGridParams or RListParams,
            reduce_factor: float = None,
            reduce_mode: str = 'nearest',
            figsize: Tuple[int, int] = None) -> None:
    if isinstance(parameters, RGridParams):
        return _plot_all_grid(generator, parameters, reduce_factor,
                              reduce_mode, figsize)
    elif isinstance(parameters, RListParams):
        return _plot_all_list(generator, parameters, reduce_factor,
                              reduce_mode, figsize)
    else:
        raise TypeError(f'Parameters have unexpected type {type(parameters)}.')


def _plot_all_grid(generator: GeneratorType, parameters: RGridParams,
                   reduce_factor: float = None,
                   reduce_mode: str = 'nearest',
                   figsize: Tuple[int, int] = None) -> None:
    nx, ny = parameters.xy_size

    figsize = figsize or (nx, ny)
    fig, axs = plt.subplots(nrows=ny, ncols=nx, sharex=True, figsize=figsize)
    axs = np.flip(axs.transpose(), axis=1).flatten()

    for i, (ax, (res, cut)) in enumerate(tqdm(zip(axs, generator))):
        if reduce_factor:
            res = zoom(res, reduce_factor, mode=reduce_mode)
        _imshow(ax, res)
        if i < ny:
            _set_y_label(ax, cut)
        if i % ny == 0:
            _set_x_label(ax, cut)

    plt.subplots_adjust(wspace=0, hspace=0)


def _plot_all_list(generator: GeneratorType, parameters: RListParams,
                   reduce_factor: float = None,
                   reduce_mode: str = 'nearest',
                   figsize: Tuple[int, int] = None) -> None:
    n = parameters.number_of_cuts
    axis_ind = parameters.axis
    if axis_ind == 0:
        nx, ny = n, 1
    else:
        nx, ny = 1, n

    figsize = figsize or (nx, ny)
    fig, axs = plt.subplots(nrows=ny, ncols=nx, figsize=figsize)

    for i, (ax, (res, cut)) in enumerate(tqdm(zip(axs, generator))):
        if reduce_factor:
            res = zoom(res, reduce_factor, mode=reduce_mode)
        _imshow(ax, res)
        if i == 0 and axis_ind == 0:
            _set_y_label(ax, cut)
        if i == len(axs) - 1 and axis_ind == 1:
            _set_x_label(ax, cut)
        if axis_ind == 0:
            _set_x_label(ax, cut)
        else:
            _set_y_label(ax, cut)

    plt.subplots_adjust(wspace=0, hspace=0)


def _imshow(ax, res) -> None:
    ax.imshow(res, cmap='jet', origin=True)
    ax.set_xticks([], [])
    ax.set_yticks([], [])


def _set_x_label(ax: plt.Axes, cut: RegionType) -> None:
    ax.set_xlabel(str((cut[0] + cut[1]) / 2)[:5])


def _set_y_label(ax: plt.Axes, cut: RegionType) -> None:
    ax.set_ylabel(str((cut[2] + cut[3]) / 2)[:5])


def _get_rect(region: RegionType, **kwargs) -> patches.Rectangle:
    return patches.Rectangle((region[0], region[2]), region[1] - region[0],
                             region[3] - region[2],
                             fill=False, color='red', **kwargs)


def _save_to_h5(name: str, folder_number: int,
                generator: GeneratorType,
                parameters: Union[RGridParams, RListParams]) -> None:
    h5_path = h5_manager.xpcs_path
    if h5_path is None:
        raise ValueError('The path to xpcs h5 file is not set.')

    with File(h5_path, 'a') as f:
        if str(folder_number) not in f.keys():
            folder = f.create_group(str(folder_number))
        else:
            folder = f[str(folder_number)]
        if name in folder:
            del folder[name]
        group = folder.create_group(name)
        group.attrs.update(parameters.dict)

    for i, (data, region) in enumerate(tqdm(generator)):
        with File(h5_path, 'a') as f:
            folder = f[str(folder_number)][name]
            dset = folder.create_dataset(str(i), data=data)
            dset.attrs['region'] = region


def _read_from_h5(folder_number: int, name: str, get_params_only: bool = False) -> GeneratorType:
    h5_path = h5_manager.xpcs_path
    if h5_path is None:
        raise ValueError('The path to xpcs h5 file is not set.')
    with File(h5_path, 'a') as f:
        if str(folder_number) not in f.keys():
            raise ValueError(f'Folder number {folder_number} is not found.')
        folder = f[str(folder_number)]
        if name not in folder.keys():
            raise ValueError(f'Group {name} is not found.')

        group = folder[name]
        if get_params_only:
            return dict(group.attrs)
        group_keys = list(group.keys())
        group_keys.sort(key=lambda x: int(x))

    def func(i: int) -> Tuple[np.ndarray, RegionType]:
        # open/close the file every time might seem a bad idea
        # however, usually it is even faster and definitely safer when using from generator
        with File(h5_path, 'a') as f_:
            group_ = f_[str(folder_number)][name]
            try:
                dset = group_[str(i)]
                region = dset.attrs.get('region', None)
                data = dset[()]
                return data, region
            except KeyError:
                raise KeyError(f'Index {i} not found.')

    yield from (func(k) for k in group_keys)


def _get_regions_from_h5(folder_number: int, name: str,
                         regions: List[RegionType]) -> Dict[RegionType, np.ndarray]:
    h5_path = h5_manager.xpcs_path
    if h5_path is None:
        raise ValueError('The path to xpcs h5 file is not set.')
    data_dict = dict()
    with File(h5_path, 'a') as f:
        if str(folder_number) not in f.keys():
            raise ValueError(f'Folder number {folder_number} is not found.')
        folder = f[str(folder_number)]
        if name not in folder.keys():
            raise ValueError(f'Group {name} is not found.')
        group = folder[name]
        for k in group.keys():
            dset = group[k]
            region = tuple(dset.attrs.get('region', None))
            if region in regions:
                data_dict[region] = dset[()]
            if len(data_dict) == len(regions):
                break
        return data_dict



# -*- coding: utf-8 -*-
from typing import Union, Tuple, NamedTuple

RegionType = Tuple[float, float, float, float]
RegionTypeShort = Tuple[float, float]


class RListParams(NamedTuple):
    init_window: Union[RegionType, RegionTypeShort]
    step: float
    axis: int
    number_of_cuts: int
    mode: str = 'list'

    @property
    def dict(self) -> dict:
        return self._asdict()


class RGridParams(NamedTuple):
    init_window: Union[RegionType, RegionTypeShort]
    xy_step: Tuple[float, float]
    xy_size: Tuple[int, int]
    mode: str = 'grid'

    @property
    def dict(self) -> dict:
        return self._asdict()


def r_params_from_dict(params: dict) -> Union[RListParams, RGridParams]:
    try:
        mode = params.pop('mode')
    except KeyError:
        raise KeyError('Mode key is missing.')
    if mode == 'list':
        try:
            return RListParams(**params)
        except TypeError:
            raise KeyError(f'Some arguments are missing: {params}')
    elif mode == 'grid':
        try:
            return RGridParams(**params)
        except TypeError:
            raise KeyError(f'Some arguments are missing: {params}')
    else:
        raise ValueError(f'Unknown mode {mode}.')


class RegionGrid(list):
    """
    RegionGrid is a list containing elements
    of type Tuple[x_min, x_max, y_min, y_max].
    x_min, x_max, y_min, y_max are values in degrees
    that define a window on a GISAXS image within which
    XPCS correlation functions will be calculated.

    Should be only initialized via classmethods and never from __init__().
    """

    __slots__ = ['parameters']

    @classmethod
    def get_cut_list(cls, init_window: Union[RegionType, RegionTypeShort],
                     axis: int, step: float, number_of_cuts: int):
        self = cls()
        if axis not in (0, 1):
            raise ValueError('Axis value should be 0 (cuts along q parallel)'
                             ' or 1 (cuts along q z axis)')
        if len(init_window) == 2:
            init_window = (init_window[0], init_window[0] + step,
                           init_window[1], init_window[1] + step)
        if axis == 0:
            self.append([(init_window[0] + step * i,
                          init_window[1] + step * i,
                          init_window[2],
                          init_window[3]) for i in range(number_of_cuts)])
        else:
            self.append([(init_window[0],
                          init_window[1],
                          init_window[2] + step * i,
                          init_window[3] + step * i)
                         for i in range(number_of_cuts)])
        setattr(self, 'parameters', RListParams(init_window=init_window, step=step,
                                                axis=axis, number_of_cuts=number_of_cuts,
                                                mode='list'))
        return self

    @classmethod
    def get_cut_grid(cls, init_window: Union[RegionType, RegionTypeShort],
                     xy_step: Tuple[float, float],
                     xy_size: Tuple[int, int]):
        if len(init_window) == 2:
            init_window = (init_window[0], init_window[0] + xy_step[0],
                           init_window[1], init_window[1] + xy_step[1])
        self = cls()
        for i in range(xy_size[0]):
            for j in range(xy_size[1]):
                self.append((init_window[0] + xy_step[0] * i,
                             init_window[1] + xy_step[0] * i,
                             init_window[2] + xy_step[1] * j,
                             init_window[3] + xy_step[1] * j))
        setattr(self, 'parameters', RGridParams(init_window=init_window, xy_step=xy_step,
                                                xy_size=xy_size, mode='grid'))
        return self

    @classmethod
    def init_from_parameters_dict(cls, parameters: Union[dict, RListParams, RGridParams]):
        if isinstance(parameters, (RListParams, RGridParams)):
            parameters = parameters.dict
        try:
            mode = parameters.pop('mode')
            if mode == 'grid':
                return cls.get_cut_grid(**parameters)
            elif mode == 'list':
                return cls.get_cut_list(**parameters)
            else:
                raise ValueError('Mode should be "grid" or "list".')

        except TypeError as err:
            raise KeyError('Provided parameters dict does not contain'
                           ' necessary parameter. Error: %s ' % err)

    def get_ranges(self) -> dict:
        def func(i):
            return [x[i] for x in self]

        x_range = (min(func(0)), max(func(1)))
        y_range = (min(func(2)), max(func(3)))
        return dict(x_range=x_range, y_range=y_range)

    def __repr__(self):
        return f'<RegionGrid: {self.parameters}\n' \
            f'Values:{super().__repr__()}>'



# -*- coding: utf-8 -*-
import re
from typing import Union, Tuple
from pathlib import Path

from h5py import File, Group
import numpy as np
from numpy import ndarray

__all__ = ['get_intensities']


def get_intensities(h5file: Union[File, Group, Path, str],
                    mask: ndarray, group_key: str = None,
                    time_key: str = 'time_of_frame') -> Tuple[ndarray, ndarray]:
    if isinstance(h5file, (File, Group)):
        return _get_intensities_from_file(h5file, mask, group_key, time_key)
    elif isinstance(h5file, str):
        path = Path(h5file)
    elif isinstance(h5file, Path):
        path = h5file
    else:
        raise TypeError(f'h5file should be Union[File, Group, Path, str] type,'
                        f'provided {type(h5file)} instead.')
    if not path.is_file():
        raise FileNotFoundError(f'File {h5file} not found.')
    with File(path, 'r') as f:
        return _get_intensities_from_file(f, mask, group_key, time_key)


_DIGITS_PATTERN = r'^([\s\d]+)$'


def _get_intensities_from_file(f: File, mask: ndarray,
                               group_key: str = None,
                               time_key: str = 'time_of_frame') -> Tuple[ndarray, ndarray]:
    if group_key:
        f = f[group_key]
    if mask.dtype != bool:
        raise TypeError(f'Mask dtype should be bool, provided {mask.dtype}')

    dset_keys = [k for k in f.keys() if re.search(_DIGITS_PATTERN, k)]
    dset_sorted_keys = sorted(dset_keys, key=lambda x: int(x))
    time_size = len(dset_sorted_keys)
    intensity_array = np.empty((time_size, np.sum(mask)))
    t_array = np.empty(time_size)

    for dset_number, dset_name in enumerate(dset_sorted_keys):
        dataset = f[dset_name]
        # if dataset.shape != mask.shape:
        #     raise ValueError(f'Mask shape is wrong: {dataset.shape} != {mask.shape}.')
        intensity = dataset[mask]
        intensity_array[dset_number, :] = intensity
        t_array[dset_number] = dataset.attrs.get(time_key, dset_number)
    return intensity_array, t_array



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from abc import abstractmethod
from typing import Tuple

import numpy as np
from scipy.ndimage.filters import gaussian_filter1d
from scipy.optimize import curve_fit

import matplotlib.pyplot as plt
import matplotlib.patches as patches
from matplotlib import rc

from .extract_data import ExtractData
from .common_tools import get_pen_folder_number


font_size = 20

rc('text', usetex=True)
rc('font', size=font_size)
rc('xtick', labelsize=font_size)
rc('ytick', labelsize=font_size)
rc('axes', labelsize=font_size)

extract_data = ExtractData()
extent = extract_data.get_extent


class Fit(object):
    init_ = ()
    args = None

    def __init__(self, fit_function):
        self.fit_function = fit_function
        self.fit_res = None
        self.fitted_y = None
        self.arg_errors = ['error_%s' % er
                           for er in self.args] if self.args is not None else None

    def fit(self, x, y, p0=None):
        p0 = p0 or self.init_
        try:
            self.fit_res = curve_fit(self.fit_function, x, y, p0)
            self.fitted_y = self.fit_function(x, *self.fit_res[0])
        except RuntimeError as er:
            print(er)

    def get_fit_result(self, mode: str = 'lists'):
        if self.fit_res is None:
            return
        if self.args is None:
            return self.fit_res
        else:
            if mode not in ['lists', 'dict']:
                raise ValueError('Mode should be one of "lists", "dict". ')
            if mode == 'dict':
                return dict(zip([*self.args, *self.arg_errors],
                                [*self.fit_res[0], *np.sqrt(np.diag(self.fit_res[1]))]))
            else:
                return self.fit_res[0], np.sqrt(np.diag(self.fit_res[1]))


class AbstractLorentzian(Fit):
    number_of_peaks = 0

    def __init__(self):
        super(AbstractLorentzian, self).__init__(self.func)

    @property
    @abstractmethod
    def init_(self):
        pass

    @init_.setter
    def init_(self, value):
        self.init_ = value

    @abstractmethod
    def func(self, *args, **kwargs):
        pass

    @staticmethod
    def single_lorentzian(x, a, w, x0):
        return a / ((x - x0) ** 2 + w)

    def init_args(self):
        return self.init_

    def fit_and_plot(self, x, y):
        self.fit(x, y)
        self.plot(x, y)

    def plot(self, x, y=None, args=None):
        if args is not None:
            pass
        elif self.fit_res is not None:
            args = self.fit_res[0]
        else:
            args = self.init_args()
        if y is not None:
            plt.semilogy(x, y, label='smoothed cut', lw=5, alpha=0.7)
        plt.semilogy(x, self.func(x, *args), label='fit')

        for i in range(self.number_of_peaks):
            plt.semilogy(x, self.single_lorentzian(x, *args[i * 3:(i + 1) * 3]), '--',
                         label=r'lorentzian %d, $x_0 = %.3f^{\circ}$' % (i + 1, args[i * 3 + 2]))
        plt.legend()


class Lorentzian2(AbstractLorentzian):
    number_of_peaks = 2

    init_ = (0.001, 0.0001, 0,
             0.0005, 0.000005, 0)

    args = ['a1', 'w1', 'x1', 'a2', 'w2', 'x2']

    def func(self, x, a1, w1, x1, a2, w2, x2):
        return self.single_lorentzian(x, a1, w1, x1) + \
               self.single_lorentzian(x, a2, w2, x2)


class Lorentzian3(AbstractLorentzian):
    number_of_peaks = 3

    init_ = (0.001, 0.0001, -0.04,
             0.001, 0.0001, 0.04,
             0.0005, 0.000005, 0)

    args = ['a1', 'w1', 'x1', 'a2', 'w2', 'x2', 'a3', 'w3', 'x3']

    def func(self, x, a1, w1, x1, a2, w2, x2, a3, w3, x3):
        return self.single_lorentzian(x, a1, w1, x1) + \
               self.single_lorentzian(x, a2, w2, x2) + \
               self.single_lorentzian(x, a3, w3, x3)


class Lorentzian4(AbstractLorentzian):
    number_of_peaks = 4

    init_ = (0.001, 0.0001, -0.04,
             0.001, 0.0001, 0.04,
             0.001, 0.0001, -0.01,
             0.001, 0.0001, 0.01)

    args = ['a1', 'w1', 'x1', 'a2', 'w2', 'x2', 'a3', 'w3', 'x3', 'a4', 'w4', 'x4']

    def func(self, x,
             a1, w1, x1,
             a2, w2, x2,
             a3, w3, x3,
             a4, w4, x4):
        return self.single_lorentzian(x, a1, w1, x1) + \
               self.single_lorentzian(x, a2, w2, x2) + \
               self.single_lorentzian(x, a3, w3, x3) + \
               self.single_lorentzian(x, a4, w4, x4)


class Lorentzian5(AbstractLorentzian):
    number_of_peaks = 5

    args = ['a1', 'w1', 'x1',
            'a2', 'w2', 'x2',
            'a3', 'w3', 'x3',
            'a4', 'w4', 'x4',
            'a5', 'w5', 'x5']

    init_ = (0.001, 0.0001, -0.04,
             0.001, 0.0001, 0.04,
             0.0005, 0.000005, 0,
             0.001, 0.0001, -0.01,
             0.001, 0.0001, 0.01)

    def func(self, x,
             a1, w1, x1,
             a2, w2, x2,
             a3, w3, x3,
             a4, w4, x4,
             a5, w5, x5):
        return self.single_lorentzian(x, a1, w1, x1) + \
               self.single_lorentzian(x, a2, w2, x2) + \
               self.single_lorentzian(x, a3, w3, x3) + \
               self.single_lorentzian(x, a4, w4, x4) + \
               self.single_lorentzian(x, a5, w5, x5)


class Lorentzian7(AbstractLorentzian):
    number_of_peaks = 7

    init_ = (0.001, 0.0001, -0.04,
             0.001, 0.0001, 0.04,
             0.0005, 0.000005, 0,
             0.001, 0.0001, -0.01,
             0.001, 0.0001, 0.01,
             0.001, 0.0001, -0.015,
             0.001, 0.0001, 0.015)

    def func(self, x,
             a1, w1, x1,
             a2, w2, x2,
             a3, w3, x3,
             a4, w4, x4,
             a5, w5, x5,
             a6, w6, x6,
             a7, w7, x7):
        return self.single_lorentzian(x, a1, w1, x1) + \
               self.single_lorentzian(x, a2, w2, x2) + \
               self.single_lorentzian(x, a3, w3, x3) + \
               self.single_lorentzian(x, a4, w4, x4) + \
               self.single_lorentzian(x, a5, w5, x5) + \
               self.single_lorentzian(x, a6, w6, x6) + \
               self.single_lorentzian(x, a7, w7, x7)


class AbstractVerticalFit(AbstractLorentzian):
    def plot(self, x, y=None, args=None):
        if args is not None:
            pass
        elif self.fit_res is not None:
            args = self.fit_res[0]
        else:
            args = self.init_args()
        if y is not None:
            plt.semilogy(x, y, label='smoothed cut', lw=5, alpha=0.7)
        plt.semilogy(x, self.func(x, *args), label='fit')

        for i in range(self.number_of_peaks):
            plt.semilogy(x, self.single_lorentzian(x, *args[i * 3:(i + 1) * 3]), '--',
                         label=r'lorentzian %d, $x_0 = %.3f^{\circ}$' % (i + 1, args[i * 3 + 2]))
        plt.semilogy(x, args[-1] * np.exp(args[-2] + x * args[-3]), '--',
                     label='exponent')
        plt.legend()


class VerticalFit2(AbstractVerticalFit):
    number_of_peaks = 2

    init_ = (0.005, 0.0001, 0.185,
             0.001, 0.0001, 0.225,
             -55, 0, 5e5)

    def func(self, x,
             a1, w1, x1,
             a2, w2, x2,
             e_tau, e_tau_0, e_amplitude):
        return e_amplitude * np.exp(e_tau_0 + x * e_tau) + \
               self.single_lorentzian(x, a1, w1, x1) + \
               self.single_lorentzian(x, a2, w2, x2)


class VerticalFit3(AbstractVerticalFit):
    number_of_peaks = 3

    init_ = (0.007, 0.0001, 0.160,
             0.005, 0.0001, 0.185,
             0.001, 0.0001, 0.225,
             -55, 0, 5e5)

    def func(self, x,
             a1, w1, x1,
             a2, w2, x2,
             a3, w3, x3,
             e_tau, e_tau_0, e_amplitude):
        return e_amplitude * np.exp(e_tau_0 + x * e_tau) + \
               self.single_lorentzian(x, a1, w1, x1) + \
               self.single_lorentzian(x, a2, w2, x2) + \
               self.single_lorentzian(x, a3, w3, x3)


class VerticalFit4(AbstractVerticalFit):
    number_of_peaks = 4

    init_ = (0.007, 0.0001, 0.160,
             0.005, 0.0001, 0.185,
             0.001, 0.0001, 0.225,
             0.0005, 0.0001, 0.280,
             -55, 0, 5e5)

    args = ['a1', 'w1', 'x1',
            'a2', 'w2', 'x2',
            'a3', 'w3', 'x3',
            'a4', 'w4', 'x4',
            'e_tau', 'e_tau_0', 'e_amplitude']

    def func(self, x,
             a1, w1, x1,
             a2, w2, x2,
             a3, w3, x3,
             a4, w4, x4,
             e_tau, e_tau_0, e_amplitude):
        return e_amplitude * np.exp(e_tau_0 + x * e_tau) + \
               self.single_lorentzian(x, a1, w1, x1) + \
               self.single_lorentzian(x, a2, w2, x2) + \
               self.single_lorentzian(x, a3, w3, x3) + \
               self.single_lorentzian(x, a4, w4, x4)


class Cut(object):
    def __init__(self, cut_region: Tuple[float, float, float, float], axis: int, *,
                 sigma: float = 1, fit: Fit = None):
        """

        :param cut_region: tuple(y_min, y_max, x_min, x_max) in deg.
        :param axis: 0 or 1 - axis number to calculate mean
        :param sigma: gaussian filter sigma (applied if sigma > 0)
        :param fit: fit object (if None, don't fit)
        """
        self.cut_region = cut_region
        self.sigma = sigma
        self.axis = axis
        self.fit = fit
        self.x = None
        self.y = None

    def set_df(self, df):
        cut = self.cut_region
        df = np.exp(df.iloc[(df.index > cut[0]) * (df.index < cut[1]),
                            (df.columns > cut[2]) * (df.columns < cut[3])])
        axes = (df.columns, df.index)
        y = df.mean(axis=self.axis)
        if self.sigma > 0:
            y = gaussian_filter1d(y, sigma=self.sigma)

        self.x, self.y = axes[self.axis], y

    def plot_cut(self, df):
        self.set_df(df)
        y = self.y

        if isinstance(self.fit, AbstractLorentzian):
            self.fit.fit_and_plot(self.x, y)
        else:
            plt.semilogy(self.x, y)
            if self.fit is not None:
                try:
                    self.plot_fit()
                except Exception as err:
                    print(err)
                    return
        ax = plt.gca()
        if self.axis == 0:
            ax.set_xlabel(r'$2 \theta$, deg')
            plt.title('Horizontal cut (smoothed)')
        else:
            ax.set_xlabel(r'$\alpha$, deg')
            plt.title('Vertical cut (smoothed)')
        ax.set_ylabel(r'$log(I)$')
        ax.grid(linestyle='--', linewidth=0.5)
        ax.set_ylim([np.min(y) * 0.95, np.max(y) * 1.05])

    def plot_fit(self):
        if self.fit is None or self.x is None:
            return
        self.fit.fit(self.x, self.y)
        if self.fit.fitted_y is None:
            return
        plt.semilogy(self.x, self.fit.fitted_y)


class PlotReport(object):
    def __init__(self, fn: int, n: int, cut_x: Cut, cut_y: Cut, *,
                 background: float = 0.1, **kwargs):
        self.fn = fn
        self.n = n
        self.cut_x = cut_x
        self.cut_y = cut_y
        self.background = background
        self.rectangle_props = dict(fill=False, color='red', alpha=1, lw=1, ls='--')
        self.rectangle_props.update(kwargs.pop('rect', dict()))
        self.df = extract_data.get_images(self.fn, self.n, apply_log=True,
                                          plot=False, background=self.background, **kwargs)

    def plot_im(self, ax=None):
        df = self.df
        if ax is None:
            plt.figure(figsize=(7, 7))
            ax = plt.gca()
        extent_ = extent(df)
        im = ax.imshow(df, extent=extent_, cmap='jet')
        ax.set_title(self.get_title())
        cbar = plt.colorbar(im, fraction=0.046, pad=0.04)
        cbar.set_label(r'$log(I)$')
        ax.set_xlabel(r'$2 \theta$, deg')
        ax.set_ylabel(r'$\alpha$, deg')

        self.add_rect(self.cut_x, ax)
        self.add_rect(self.cut_y, ax)

    def plot(self, figsize: Tuple[int, int] = (5, 15)):
        self.plot_im()
        df = self.df
        plt.figure(figsize=figsize)
        plt.subplot(211)
        self.cut_x.plot_cut(df)
        plt.subplot(212)
        self.cut_y.plot_cut(df)
        plt.tight_layout()

    def get_title(self):
        fn, n = self.fn, self.n
        title = 'Pentacene sample %d' % get_pen_folder_number(fn)
        try:
            title = r'%s%s \textit{(folder %d, image %d)}' % (
                title,
                extract_data.metadata.get_latex_title(fn, n),
                fn, n)
        except Exception as err:
            print(err)
        return title

    def add_rect(self, cut_obj: Cut, ax):
        if cut_obj is not None:
            cut = cut_obj.cut_region
            r = patches.Rectangle((cut[2], cut[0]), cut[3] - cut[2], cut[1] - cut[0], **self.rectangle_props)
            ax.add_patch(r)


if __name__ == '__main__':
    from pprint import pprint

    lfit2 = Lorentzian2()
    lfit3 = Lorentzian3()
    lfit4 = Lorentzian4()
    lfit5 = Lorentzian5()
    lfit7 = Lorentzian7()
    vfit2 = VerticalFit2()
    vfit3 = VerticalFit3()
    vfit4 = VerticalFit4()

    # fit for 99 - 1500 (low central peak)

    lfit5.init_ = (0.0005, 0.0001, -0.04,
                   0.0005, 0.0001, 0.04,
                   0.005, 0.001, 0,
                   0.0007, 0.00002, -0.005,
                   0.0007, 0.00002, 0.005)

    # fit for 97 - 1 (low central peak)

    # lfit5.init_ = (0.0005, 0.0001, -0.04,
    #                0.0005, 0.0001, 0.04,
    #                0.001, 0.001, 0,
    #                0.0003, 0.00002, -0.005,
    #                0.0003, 0.00002, 0.005)

    # for two central peaks 96 - 1:

    # lfit4.init_ = [2.22402770e-03, 3.22734594e-04, - 3.58788410e-02,
    #                3.27324732e-03, 4.52631052e-04, 3.70929532e-02,
    #                2.66907060e-04, 5.16160062e-06, 1.44646801e-03,
    #                1.85142830e-03, 3.41261952e-04, 0]

    # cut_x = Cut((0.18, 0.20, -0.07, 0.07), 0, sigma=2, fit=lfit3)

    #####
    # cut for 89

    # vfit2.init_ = (0.007, 0.0001, 0.160,
    #                0.001, 0.0001, 0.234,
    #                -55, 0, 5e5)
    # cut_x = Cut((0.22, 0.24, -0.07, 0.07), 0, sigma=2, fit=lfit3)
    # cut_y = Cut((0.15, 0.3, -0.05, -0.03), 1, sigma=1, fit=vfit2)
    #####


    def get_nm(delta, name=''):
        print(name, ':\n', 8.77 / delta, ' nm')


    cut_x = Cut((0.22, 0.24, -0.07, 0.07), 0, sigma=1, fit=lfit5)
    cut_y = Cut((0.15, 0.3, -0.01, 0.01), 1, sigma=2, fit=vfit4)
    # cut_y = Cut((0.15, 0.3, -0.01, 0.01), 1, sigma=2
    plot = PlotReport(99, 1, cut_x=cut_x, cut_y=cut_y)
    plot.plot(figsize=(10, 10))
    plt.show()
    dh = lfit5.get_fit_result()
    dv = vfit4.get_fit_result()

    get_nm((dh['x5'] - dh['x4']) / 2, 'big x')
    get_nm((dh['x2'] - dh['x1']) / 2, 'usual x')
    get_nm(dv['x2'] - dv['x1'], 'y1')
    get_nm(dv['x3'] - dv['x2'], 'y2')
    get_nm(dv['x4'] - dv['x3'], 'y3')

    # for i in [500, 550, 600, 650, 700, 750]:
    #
    #     plot = PlotReport(87, i, cut_x=cut_x, cut_y=cut_y)
    #     plt.figure(figsize=(10, 10))
    #     plot.plot_im()
    #     plt.savefig('images_for_report/87_%d.png' % i)
    #     plt.figure(figsize=(10, 5))
    #     cut_x.plot_cut(plot.df)
    #     # plot.plot(figsize=(10, 10))
    #     plt.savefig('images_for_report/87_%d_hcut.png' % i)

    # print(list(lfit5.get_fit_result()))



# -*- coding: utf-8 -*-
import pytest
import numpy as np

from gisaxs_xpcs import XPCS


def test_get_intensities(folder, region):
    xpcs = XPCS(folder, region)
    intensity, t_array = xpcs.get_intensities()
    assert intensity.shape[0] == t_array.size


def test_mask(folder):
    region = (-10, 10, -10, 10)
    xpcs = XPCS(folder, region)
    mask_size = xpcs.mask.shape[0] * xpcs.mask.shape[1]
    assert np.sum(xpcs.mask) == mask_size


def test_2d_corr(folder, region):
    xpcs = XPCS(folder, region)
    corr, t_array = xpcs.get_two_time_corr_function()
    assert corr.shape == (t_array.size, t_array.size)



# -*- coding: utf-8 -*-

from enum import Enum, auto
import numpy as np


class XPCSModes(Enum):
    mean = auto()
    std = auto()


def calculate_2d_corr(intensity_array: np.ndarray,
                      mode: XPCSModes) -> np.ndarray:
    """
    Calculated two-time correlation function from an intensity array
    for two possible modes: mean or std.
    :param intensity_array: A 2-dimensional ndarray with a shape (time_axis, pixel_axis).
    :param mode: a XPCSModes value (XPCSModes.mean or XPCSModes.std).
    :return: A 2-dimensional array with a shape (time_axis, time_axis).
    """
    if mode == XPCSModes.mean:
        intensity_array /= intensity_array.mean(axis=1)[:, np.newaxis]
        return intensity_array.dot(intensity_array.transpose()) / intensity_array.shape[1]
    elif mode == XPCSModes.std:
        std_array = intensity_array.std(axis=1)[:, np.newaxis]
        std_array[std_array == 0] = 1
        intensity_array /= std_array
        res = intensity_array.dot(intensity_array.transpose()) / intensity_array.shape[1]
        i_mean = np.expand_dims(intensity_array.mean(axis=1), axis=1)
        res -= i_mean.dot(i_mean.transpose())
        return res
    else:
        raise ValueError(f'Unknown XPCS mode: {mode}.')


def calculate_1d_corr(intensity_array: np.ndarray) -> np.ndarray:
    """
    Calculates one-time correlation function from an intensity array.
    :param intensity_array: A 2-dimensional ndarray with a shape (time_axis, pixel_axis).
    :return: A 1-dimensional array with a size time_axis.
    """

    norm_constant = intensity_array.mean(axis=1).mean(axis=0) ** 2
    corr_matrix = intensity_array.dot(intensity_array.transpose()) / intensity_array.shape[1]
    corr_function = np.empty(intensity_array.shape[0])

    for t in range(corr_matrix.shape[0]):
        corr_function[t] = corr_matrix.diagonal(t).mean() / norm_constant
    return corr_function
